#include "BuddhaMatrix.h"
#include "BuddhaVector3.h"

namespace BuddhaXNA
{
	const Matrix Matrix::Identity(
							  1, 0, 0, 0,
							  0, 1, 0, 0,
							  0, 0, 1, 0,
							  0, 0, 0, 1);
	//------------------------------------------------------------------------
	Matrix Matrix::CreateLookAt(const Vector3 &from, const Vector3 &to, const Vector3 &up)
	{
		D3DXMATRIX d_mat;
		D3DXVECTOR3 d_from, d_to, d_up;
		
		Vector3::CreateD3DVECTOR(from, d_from);
		Vector3::CreateD3DVECTOR(to, d_to);
		Vector3::CreateD3DVECTOR(up, d_up);

		D3DXMatrixLookAtRH(&d_mat, &d_from, &d_to, &d_up);

		Matrix mat;
		
		mat = d_mat;

		return mat;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreatePerspectiveFieldOfView(FLOAT fieldOfView, FLOAT aspectRatio, FLOAT nearPlaneDistance, FLOAT farPlaneDistance)
	{
		Matrix result;
		CreatePerspectiveFieldOfView(fieldOfView, aspectRatio, nearPlaneDistance, farPlaneDistance, result);

		return result;
	}
	//------------------------------------------------------------------------
	void Matrix::CreatePerspectiveFieldOfView(FLOAT fieldOfView, FLOAT aspectRatio, FLOAT nearPlaneDistance, FLOAT farPlaneDistance, Matrix &result)
	{
		D3DXMATRIX d_mat;
		D3DXMatrixPerspectiveFovRH(&d_mat, fieldOfView, aspectRatio, nearPlaneDistance, farPlaneDistance);
		Matrix::CreateFromD3DXMatrix(d_mat, result);
	}
	//------------------------------------------------------------------------
	void Matrix::CreateRotationX(FLOAT radians, Matrix &out)
	{
		out.M11 = 1; out.M12 = 0;                         out.M13 = 0;                          out.M14 = 0;
		out.M21 = 0; out.M22 = (FLOAT)Math::Cos(radians); out.M23 = (FLOAT)-Math::Sin(radians); out.M24 = 0;
		out.M31 = 0; out.M32 = (FLOAT)Math::Sin(radians); out.M33 = (FLOAT)Math::Cos(radians);  out.M34 = 0;
		out.M41 = 0; out.M42 = 0;                         out.M43 = 0;                          out.M44 = 1;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateRotationX(FLOAT radians)
	{
		Matrix mat;
		CreateRotationX(radians, mat);
		return mat;
	}
	//------------------------------------------------------------------------
	void Matrix::CreateRotationY(FLOAT radians, Matrix &out)
	{
		out.M11 = (FLOAT)Math::Cos(radians);  out.M12 = 0; out.M13 = (FLOAT)Math::Sin(radians);  out.M14 = 0;
		out.M21 = 0;                          out.M22 = 1; out.M23 = 0;                          out.M24 = 0;
		out.M31 = (FLOAT)-Math::Sin(radians); out.M32 = 0; out.M33 = (FLOAT)Math::Cos(radians);  out.M34 = 0;
		out.M41 = 0;                          out.M42 = 0; out.M43 = 0;                          out.M44 = 1;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateRotationY(FLOAT radians)
	{
		Matrix mat;
		CreateRotationY(radians, mat);
		return mat;
	}
	//------------------------------------------------------------------------
	void Matrix::CreateRotationZ(FLOAT radians, Matrix &out)
	{
		out.M11 = (FLOAT)Math::Cos(radians);  out.M12 = (FLOAT)-Math::Sin(radians); out.M13 = 0; out.M14 = 0;
		out.M21 = (FLOAT)Math::Sin(radians);  out.M22 = (FLOAT)Math::Cos(radians);  out.M23 = 0; out.M24 = 0;
		out.M31 = 0;                          out.M32 = 0;                          out.M33 = 1; out.M34 = 0;
		out.M41 = 0;                          out.M42 = 0;                          out.M43 = 0; out.M44 = 1;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateRotationZ(FLOAT radians)
	{
		Matrix mat;
		CreateRotationZ(radians, mat);
		return mat;
	}
	//------------------------------------------------------------------------
	void Matrix::CreateScale(const FLOAT sx, const FLOAT sy, const FLOAT sz, Matrix &out)
	{
		out.M11 = sx; out.M12 = 0;  out.M13 = 0;  out.M14 = 0;
		out.M21 = 0;  out.M22 = sy; out.M23 = 0;  out.M24 = 0;
		out.M31 = 0;  out.M32 = 0;  out.M33 = sz; out.M34 = 0;
		out.M41 = 0;  out.M42 = 0;  out.M43 = 0;  out.M44 = 1;
	}
	//------------------------------------------------------------------------
	void Matrix::CreateScale(const FLOAT scale, Matrix &out)
	{
		CreateScale(scale, scale, scale, out);
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateScale(const FLOAT sx, const FLOAT sy, const FLOAT sz)
	{
		Matrix out;
		CreateScale(sx, sy, sz, out);
		return out;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateScale(const FLOAT scale)
	{
		Matrix out;
		CreateScale(scale, scale, scale, out);
		return out;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateScale(const Vector3 &vec)
	{
		return CreateScale(vec.X, vec.Y, vec.Z);
	}
	//------------------------------------------------------------------------
	void Matrix::CreateTransform(const FLOAT x, const FLOAT y, const FLOAT z, BuddhaXNA::Matrix &out)
	{
		out.M11 = 1;  out.M12 = 0;  out.M13 = 0;  out.M14 = 0;
		out.M21 = 0;  out.M22 = 1;  out.M23 = 0;  out.M24 = 0;
		out.M31 = 0;  out.M32 = 0;  out.M33 = 1;  out.M34 = 0;
		out.M41 = x;  out.M42 = y;  out.M43 = z;  out.M44 = 1;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateTransform(const FLOAT x, const FLOAT y, const FLOAT z)
	{
		Matrix out;
		CreateTransform(x, y, z, out);
		return out;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateTransform(const Vector3 &vec)
	{
		return CreateTransform(vec.X, vec.Y, vec.Z);
	}
	//------------------------------------------------------------------------
	void Matrix::Multiply(const Matrix &mat1, const Matrix &mat2, Matrix &result)
	{
		result.M11 = mat1.M11 * mat2.M11 + mat1.M12 * mat2.M21 + mat1.M13 * mat2.M31 + mat1.M14 * mat2.M41; 
		result.M12 = mat1.M11 * mat2.M12 + mat1.M12 * mat2.M22 + mat1.M13 * mat2.M32 + mat1.M14 * mat2.M42; 
		result.M13 = mat1.M11 * mat2.M13 + mat1.M12 * mat2.M23 + mat1.M13 * mat2.M33 + mat1.M14 * mat2.M43; 
		result.M14 = mat1.M11 * mat2.M14 + mat1.M12 * mat2.M24 + mat1.M13 * mat2.M34 + mat1.M14 * mat2.M44; 

		result.M21 = mat1.M21 * mat2.M11 + mat1.M22 * mat2.M21 + mat1.M23 * mat2.M31 + mat1.M24 * mat2.M41; 
		result.M22 = mat1.M21 * mat2.M12 + mat1.M22 * mat2.M22 + mat1.M23 * mat2.M32 + mat1.M24 * mat2.M42; 
		result.M23 = mat1.M21 * mat2.M13 + mat1.M22 * mat2.M23 + mat1.M23 * mat2.M33 + mat1.M24 * mat2.M43; 
		result.M24 = mat1.M21 * mat2.M14 + mat1.M22 * mat2.M24 + mat1.M23 * mat2.M34 + mat1.M24 * mat2.M44; 

		result.M31 = mat1.M31 * mat2.M11 + mat1.M32 * mat2.M21 + mat1.M33 * mat2.M31 + mat1.M34 * mat2.M41; 
		result.M32 = mat1.M31 * mat2.M12 + mat1.M32 * mat2.M22 + mat1.M33 * mat2.M32 + mat1.M34 * mat2.M42; 
		result.M33 = mat1.M31 * mat2.M13 + mat1.M32 * mat2.M23 + mat1.M33 * mat2.M33 + mat1.M34 * mat2.M43; 
		result.M34 = mat1.M31 * mat2.M14 + mat1.M32 * mat2.M24 + mat1.M33 * mat2.M34 + mat1.M34 * mat2.M44; 

		result.M41 = mat1.M41 * mat2.M11 + mat1.M42 * mat2.M21 + mat1.M43 * mat2.M31 + mat1.M44 * mat2.M41; 
		result.M42 = mat1.M41 * mat2.M12 + mat1.M42 * mat2.M22 + mat1.M43 * mat2.M32 + mat1.M44 * mat2.M42; 
		result.M43 = mat1.M41 * mat2.M13 + mat1.M42 * mat2.M23 + mat1.M43 * mat2.M33 + mat1.M44 * mat2.M43; 
		result.M44 = mat1.M41 * mat2.M14 + mat1.M42 * mat2.M24 + mat1.M43 * mat2.M34 + mat1.M44 * mat2.M44; 
	}
	//------------------------------------------------------------------------
	Matrix Matrix::Multiply(const Matrix &mat1, const Matrix &mat2)
	{
		Matrix out;
		Multiply(mat1, mat2, out);
		return out;
	}
	//------------------------------------------------------------------------
	void Matrix::CreateFromQuaternion(const Quaternion &q, Matrix &out)
	{
		double fTx = 2.0 * q.X;
		double fTy = 2.0 * q.Y;
		double fTz = 2.0 * q.Z;

		double fTwx = fTx * q.W;
		double fTwy = fTy * q.W;
		double fTwz = fTz * q.W;
		double fTxx = fTx * q.X;
		double fTxy = fTy * q.X;
		double fTxz = fTz * q.X;

		double fTyy = fTy * q.Y;
		double fTyz = fTz * q.Y;
		double fTzz = fTz * q.Z;

		out.M11 = (FLOAT)(1.0 - (fTyy + fTzz));
		out.M12 = (FLOAT)(fTxy - fTwz);
		out.M13 = (FLOAT)(fTxz + fTwy);
		out.M14 = 0.0f;

		out.M21 = (FLOAT)(fTxy + fTwz);
		out.M22 = (FLOAT)(1.0 - (fTxx + fTzz));
		out.M23 = (FLOAT)(fTyz - fTwx);
		out.M24 = 0.0f;

		out.M31 = (FLOAT)(fTxz - fTwy);
		out.M32 = (FLOAT)(fTyz + fTwx);
		out.M33 = (FLOAT)(1.0f - (fTxx + fTyy));
		out.M34 = 0.0f;

		out.M41 = out.M42 = out.M43 = 0.0f;
		out.M44 = 1.0f;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateFromQuaternion(const Quaternion &q)
	{
		Matrix ret;
		CreateFromQuaternion(q, ret);
		return ret;
	}
	//------------------------------------------------------------------------
	Matrix Matrix::CreateBillboard(const Vector3 &objectPosition, const Vector3 &cameraPosition,
			const Vector3 &cameraUpVector)
	{
		Matrix result;
		CreateBillboard(objectPosition, cameraPosition, cameraUpVector, result);

		return result;
	}
	//------------------------------------------------------------------------
	void Matrix::CreateBillboard(const Vector3 &objectPosition, const Vector3 &cameraPosition,
			const Vector3 &cameraUpVector, Matrix &result)
	{
		D3DXMATRIX dxMat;
		D3DXVECTOR3 dxFrom, dxTo, dxUp;
		
		Vector3::CreateD3DVECTOR(Vector3::Zero, dxFrom);
		Vector3::CreateD3DVECTOR(objectPosition - cameraPosition, dxTo);
		Vector3::CreateD3DVECTOR(cameraUpVector, dxUp);

		D3DXMatrixIdentity(&dxMat);

		// ]ŝ݂쐬
		D3DXMatrixLookAtRH(&dxMat, &dxFrom, &dxTo, &dxUp);

		// ts쐬
		D3DXMatrixInverse(&dxMat, NULL, &dxMat);

		// 쐬s
		Matrix::CreateFromD3DXMatrix(dxMat, result);
	}
}
