#include "BuddhaVector3.h"
#include "BuddhaMatrix.h"
#include "BuddhaQuaternion.h"

#include "BuddhaMath.h"

namespace BuddhaXNA
{
	const Vector3 Vector3::Zero(0.0f, 0.0f, 0.0f);
	const Vector3 Vector3::One(1.0f, 1.0f, 1.0f);

	const Vector3 Vector3::Left(-1.0f, 0.0f, 0.0f);
	const Vector3 Vector3::Right(1.0f, 0.0f, 0.0f);

	const Vector3 Vector3::Up(0.0f, 1.0f, 0.0f);
	const Vector3 Vector3::Down(0.0f, -1.0f, 0.0f);

	const Vector3 Vector3::Forward(0.0f, 0.0f, -1.0f);
	const Vector3 Vector3::Backward(0.0f, 0.0f, 1.0f);

	const Vector3 Vector3::UnitX(1.0f, 0.0f, 0.0f);
	const Vector3 Vector3::UnitY(0.0f, 1.0f, 0.0f);
	const Vector3 Vector3::UnitZ(0.0f, 0.0f, 1.0f);
	//------------------------------------------------------------------------
	void Vector3::Normalize()
	{
		Real length = BuddhaXNA::Math::Sqrt(X * X + Y * Y + Z * Z);

		if (length > 1e-08)
		{
			FLOAT inv_length = (FLOAT)(1.0 / length);
			X *= inv_length;
			Y *= inv_length;
			Z *= inv_length;
		}
	}
	//------------------------------------------------------------------------
	Vector3 Vector3::Negate(const Vector3& value)
	{
		return Vector3(-value.X, -value.Y, -value.Z);
	}
	//------------------------------------------------------------------------
	void Vector3::Negate(const Vector3& in, Vector3& out)
	{
		out.X = -in.X;
		out.Y = -in.Y;
		out.Z = -in.Z;
	}
	//------------------------------------------------------------------------
	FLOAT Vector3::Dot(const Vector3& value1, const Vector3& value2)
	{
		return value1.X * value2.X + value1.Y * value2.Y + value1.Z * value2.Z;
	}
	//------------------------------------------------------------------------
	Vector3 Vector3::Cross(const Vector3& vector1, const Vector3& vector2)
	{
		return Vector3(
			vector1.Y * vector2.Z - vector1.Z * vector2.Y,
			vector1.Z * vector2.X - vector1.X * vector2.Z,
			vector1.X * vector2.Y - vector1.Y * vector2.X);
	}
	//------------------------------------------------------------------------
	Vector3 Vector3::Reflect(const Vector3& value, const Vector3& normal)
	{
		return Vector3(value - (2.0f * Vector3::Dot(value, normal) * normal));
	}
	//------------------------------------------------------------------------
	void Vector3::Reflect(const Vector3& value, const Vector3& normal, Vector3& out)
	{
		out = value - (2.0f * Vector3::Dot(value, normal) * normal);
	}
	//------------------------------------------------------------------------
	Vector3 Vector3::Transform(const Vector3& vector, const Matrix& matrix)
	{
		Vector3 ret;
		FLOAT t_invW = (FLOAT)(1.0 / (matrix.M41 * vector.X + matrix.M42 * vector.Y + matrix.M43 * vector.Z + matrix.M44));

		ret.X = (matrix.M11 * vector.X + matrix.M12 * vector.Y + matrix.M13 * vector.Z) * t_invW;
		ret.Y = (matrix.M21 * vector.X + matrix.M22 * vector.Y + matrix.M23 * vector.Z) * t_invW;
		ret.Z = (matrix.M31 * vector.X + matrix.M32 * vector.Y + matrix.M33 * vector.Z) * t_invW;

		return ret;
	}
	//------------------------------------------------------------------------
	void Vector3::Transform(const Vector3& vector, const Matrix& matrix, Vector3& out)
	{
		FLOAT t_invW = (FLOAT)(1.0 / (matrix.M41 * vector.X + matrix.M42 * vector.Y + matrix.M43 * vector.Z + matrix.M44));

		out.X = (matrix.M11 * vector.X + matrix.M12 * vector.Y + matrix.M13 * vector.Z) * t_invW;
		out.Y = (matrix.M21 * vector.X + matrix.M22 * vector.Y + matrix.M23 * vector.Z) * t_invW;
		out.Z = (matrix.M31 * vector.X + matrix.M32 * vector.Y + matrix.M33 * vector.Z) * t_invW;
	}
	//------------------------------------------------------------------------
	Vector3 Vector3::Transform(const Vector3 &vector, const Quaternion &q)
	{
		Vector3 out;
		Transform(vector, Matrix::CreateFromQuaternion(q), out);
		return out;
	}
	//------------------------------------------------------------------------
	void Vector3::Transform(const Vector3 &vector, const Quaternion &q, Vector3 &out)
	{
		Matrix rotationMatrix = Matrix::CreateFromQuaternion(q);
		Transform(vector, q, out);
	}
	//------------------------------------------------------------------------
	void Vector3::CreateD3DVECTOR(const Vector3 &in, D3DVECTOR &out)
	{
		out.x = in.X;
		out.y = in.Y;
		out.z = in.Z;
	}
}

